=== modified file 'NEWS'
--- NEWS	2009-01-15 05:18:22 +0000
+++ NEWS	2009-01-15 07:34:16 +0000
@@ -4,6 +4,13 @@
 
 .. contents::
 
+IN DEVELOPMENT
+--------------
+
+  IMPROVEMENTS:
+
+    * Progress bars now show the rate of activity for some sftp 
+      operations, and they are drawn different.  (Martin Pool, #172741)
 
 NOT RELEASED YET
 ----------------
@@ -271,6 +278,10 @@
       advantage of pycurl is that it checks ssl certificates.)
       (John Arbash Meinel)
 
+    * The progress and UI classes have changed; the main APIs remain the
+      same but code that provides a new UI or progress bar class may
+      need to be updated.  (Martin Pool)
+
     * ``VersionedFiles.get_record_stream()`` can now return objects with a
       storage_kind of ``chunked``. This is a collection (list/tuple) of
       strings. You can use ``osutils.chunks_to_lines()`` to turn them into

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2008-10-21 20:43:05 +0000
+++ bzrlib/commands.py	2008-12-16 07:56:29 +0000
@@ -868,8 +868,8 @@
 
 def main(argv):
     import bzrlib.ui
-    from bzrlib.ui.text import TextUIFactory
-    bzrlib.ui.ui_factory = TextUIFactory()
+    bzrlib.ui.ui_factory = bzrlib.ui.make_ui_for_terminal(
+        sys.stdin, sys.stdout, sys.stderr)
 
     # Is this a final release version? If so, we should suppress warnings
     if bzrlib.version_info[3] == 'final':

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2009-01-08 16:57:10 +0000
+++ bzrlib/errors.py	2009-01-13 05:16:26 +0000
@@ -2171,11 +2171,6 @@
     _fmt = "Cannot perform local-only commits on unbound branches."
 
 
-class MissingProgressBarFinish(BzrError):
-
-    _fmt = "A nested progress bar was not 'finished' correctly."
-
-
 class InvalidProgressBarType(BzrError):
 
     _fmt = ("Environment variable BZR_PROGRESS_BAR='%(bar_type)s"

=== modified file 'bzrlib/progress.py'
--- bzrlib/progress.py	2008-04-24 07:22:53 +0000
+++ bzrlib/progress.py	2008-12-17 08:18:14 +0000
@@ -1,5 +1,4 @@
-# Copyright (C) 2005 Aaron Bentley <aaron.bentley@utoronto.ca>
-# Copyright (C) 2005, 2006 Canonical Ltd
+# Copyright (C) 2005, 2006, 2008 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -28,20 +27,19 @@
 bzrlib it really is best to use bzrlib.ui.ui_factory.
 """
 
-# TODO: Optionally show elapsed time instead/as well as ETA; nicer
-# when the rate is unpredictable
 
 import sys
 import time
 import os
-
-from bzrlib.lazy_import import lazy_import
-lazy_import(globals(), """
+import warnings
+
+
 from bzrlib import (
     errors,
+    osutils,
+    trace,
+    ui,
     )
-""")
-
 from bzrlib.trace import mutter
 
 
@@ -62,7 +60,69 @@
     return True
 
 
-_progress_bar_types = {}
+class ProgressTask(object):
+    """Model component of a progress indicator.
+
+    Most code that needs to indicate progress should update one of these, 
+    and it will in turn update the display, if one is present.
+
+    Code updating the task may also set fields as hints about how to display
+    it: show_pct, show_spinner, show_eta, show_count, show_bar.  UIs
+    will not necessarily respect all these fields.
+    """
+
+    def __init__(self, parent_task=None, ui_factory=None):
+        self._parent_task = parent_task
+        self._last_update = 0
+        self.total_cnt = None
+        self.current_cnt = None
+        self.msg = ''
+        self.ui_factory = ui_factory
+        self.show_pct = False
+        self.show_spinner = True
+        self.show_eta = False,
+        self.show_count = True
+        self.show_bar = True
+
+    def update(self, msg, current_cnt=None, total_cnt=None):
+        self.msg = msg
+        self.current_cnt = current_cnt
+        if total_cnt:
+            self.total_cnt = total_cnt
+        self.ui_factory.show_progress(self)
+
+    def tick(self):
+        self.update(self.msg)
+
+    def finished(self):
+        self.ui_factory.progress_finished(self)
+
+    def make_sub_task(self):
+        return ProgressTask(self, self.ui_factory)
+
+    def _overall_completion_fraction(self, child_fraction=0.0):
+        """Return fractional completion of this task and its parents
+        
+        Returns None if no completion can be computed."""
+        if self.total_cnt:
+            own_fraction = (float(self.current_cnt) + child_fraction) / self.total_cnt
+        else:
+            own_fraction = None
+        if self._parent_task is None:
+            return own_fraction
+        else:
+            if own_fraction is None:
+                own_fraction = 0.0
+            return self._parent_task._overall_completion_fraction(own_fraction)
+
+    def note(self, fmt_string, *args):
+        """Record a note without disrupting the progress bar."""
+        # XXX: shouldn't be here; put it in mutter or the ui instead
+        self.ui_factory.note(fmt_string % args)
+
+    def clear(self):
+        # XXX: shouldn't be here; put it in mutter or the ui instead
+        self.ui_factory.clear_term()
 
 
 def ProgressBar(to_file=None, **kwargs):
@@ -87,7 +147,7 @@
                                                 _progress_bar_types.keys())
         return _progress_bar_types[requested_bar_type](to_file=to_file, **kwargs)
 
- 
+
 class ProgressBarStack(object):
     """A stack of progress bars."""
 
@@ -147,8 +207,9 @@
     def return_pb(self, bar):
         """Return bar after its been used."""
         if bar is not self._stack[-1]:
-            raise errors.MissingProgressBarFinish()
-        self._stack.pop()
+            warnings.warn("%r is not currently active" % (bar,))
+        else:
+            self._stack.pop()
 
  
 class _BaseProgressBar(object):
@@ -206,6 +267,7 @@
 
     This can be used as the default argument for methods that
     take an optional progress indicator."""
+
     def tick(self):
         pass
 
@@ -225,10 +287,6 @@
         return DummyProgress(**kwargs)
 
 
-_progress_bar_types['dummy'] = DummyProgress
-_progress_bar_types['none'] = DummyProgress
-
-
 class DotsProgressBar(_BaseProgressBar):
 
     def __init__(self, **kwargs):
@@ -257,7 +315,6 @@
         self.tick()
 
 
-_progress_bar_types['dots'] = DotsProgressBar
 
     
 class TTYProgressBar(_BaseProgressBar):
@@ -329,8 +386,9 @@
         self.tick()
 
     def update(self, msg, current_cnt=None, total_cnt=None,
-               child_fraction=0):
-        """Update and redraw progress bar."""
+            child_fraction=0):
+        """Update and redraw progress bar.
+        """
         if msg is None:
             msg = self.last_msg
 
@@ -356,6 +414,9 @@
         ##     self.child_fraction == child_fraction):
         ##     return
 
+        if msg is None:
+            msg = ''
+
         old_msg = self.last_msg
         # save these for the tick() function
         self.last_msg = msg
@@ -401,7 +462,7 @@
             # make both fields the same size
             t = '%i' % (self.last_total)
             c = '%*i' % (len(t), self.last_cnt)
-            count_str = ' ' + c + '/' + t 
+            count_str = ' ' + c + '/' + t
 
         if self.show_bar:
             # progress bar, if present, soaks up all remaining space
@@ -425,7 +486,8 @@
         else:
             bar_str = ''
 
-        m = spin_str + bar_str + self.last_msg + count_str + pct_str + eta_str
+        m = spin_str + bar_str + self.last_msg + count_str \
+            + pct_str + eta_str
         self.to_file.write('\r%-*.*s' % (self.width - 1, self.width - 1, m))
         self._have_output = True
         #self.to_file.flush()
@@ -437,7 +499,6 @@
         #self.to_file.flush()        
 
 
-_progress_bar_types['tty'] = TTYProgressBar
 
 
 class ChildProgress(_BaseProgressBar):
@@ -557,3 +618,10 @@
         else:
             self.cur_phase += 1
         self.pb.update(self.message, self.cur_phase, self.total)
+
+
+_progress_bar_types = {}
+_progress_bar_types['dummy'] = DummyProgress
+_progress_bar_types['none'] = DummyProgress
+_progress_bar_types['tty'] = TTYProgressBar
+_progress_bar_types['dots'] = DotsProgressBar

=== modified file 'bzrlib/tests/blackbox/test_upgrade.py'
--- bzrlib/tests/blackbox/test_upgrade.py	2008-03-14 21:31:15 +0000
+++ bzrlib/tests/blackbox/test_upgrade.py	2009-01-13 05:07:27 +0000
@@ -1,6 +1,4 @@
-# Copyright (C) 2006, 2007 Canonical Ltd
-# Authors: Robert Collins <robert.collins@canonical.com>
-#          and others
+# Copyright (C) 2006, 2007, 2009 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by

=== modified file 'bzrlib/tests/test_sftp_transport.py'
--- bzrlib/tests/test_sftp_transport.py	2008-12-09 14:04:01 +0000
+++ bzrlib/tests/test_sftp_transport.py	2009-01-15 05:12:49 +0000
@@ -474,12 +474,17 @@
             yield self._data[start:start+length]
 
 
+def _null_report_activity(*a, **k):
+    pass
+
+
 class Test_SFTPReadvHelper(tests.TestCase):
 
     def checkGetRequests(self, expected_requests, offsets):
         if not paramiko_loaded:
             raise TestSkipped('you must have paramiko to run this test')
-        helper = _mod_sftp._SFTPReadvHelper(offsets, 'artificial_test')
+        helper = _mod_sftp._SFTPReadvHelper(offsets, 'artificial_test',
+            _null_report_activity)
         self.assertEqual(expected_requests, helper._get_requests())
 
     def test__get_requests(self):
@@ -499,7 +504,8 @@
     def checkRequestAndYield(self, expected, data, offsets):
         if not paramiko_loaded:
             raise TestSkipped('you must have paramiko to run this test')
-        helper = _mod_sftp._SFTPReadvHelper(offsets, 'artificial_test')
+        helper = _mod_sftp._SFTPReadvHelper(offsets, 'artificial_test',
+            _null_report_activity)
         data_f = ReadvFile(data)
         result = list(helper.request_and_yield_offsets(data_f))
         self.assertEqual(expected, result)

=== modified file 'bzrlib/tests/test_ui.py'
--- bzrlib/tests/test_ui.py	2007-04-13 16:16:54 +0000
+++ bzrlib/tests/test_ui.py	2009-01-15 06:44:09 +0000
@@ -1,4 +1,4 @@
-# Copyright (C) 2005 Canonical Ltd
+# Copyright (C) 2005, 2008 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -101,11 +101,11 @@
     def test_progress_note(self):
         stderr = StringIO()
         stdout = StringIO()
-        ui_factory = TextUIFactory(bar_type=TTYProgressBar)
+        ui_factory = TextUIFactory(stdin=StringIO(''),
+            stderr=stderr,
+            stdout=stdout)
         pb = ui_factory.nested_progress_bar()
         try:
-            pb.to_messages_file = stdout
-            ui_factory._progress_bar_stack.bottom().to_file = stderr
             result = pb.note('t')
             self.assertEqual(None, result)
             self.assertEqual("t\n", stdout.getvalue())
@@ -122,13 +122,12 @@
         # The PQM redirects the output to a file, so it
         # defaults to creating a Dots progress bar. we
         # need to force it to believe we are a TTY
-        ui_factory = TextUIFactory(bar_type=TTYProgressBar)
+        ui_factory = TextUIFactory(
+            stdin=StringIO(''),
+            stdout=stdout, stderr=stderr)
         pb = ui_factory.nested_progress_bar()
         try:
-            pb.to_messages_file = stdout
-            ui_factory._progress_bar_stack.bottom().to_file = stderr
             # Create a progress update that isn't throttled
-            pb.start_time -= 10
             pb.update('x', 1, 1)
             result = pb.note('t')
             self.assertEqual(None, result)
@@ -142,10 +141,12 @@
 
     def test_progress_nested(self):
         # test factory based nested and popping.
-        ui = TextUIFactory()
+        ui = TextUIFactory(None, None, None)
         pb1 = ui.nested_progress_bar()
         pb2 = ui.nested_progress_bar()
-        self.assertRaises(errors.MissingProgressBarFinish, pb1.finished)
+        # We no longer warn about finishing unnested progress bars.
+        warnings, _ = self.callCatchWarnings(pb1.finished)
+        self.assertEqual(len(warnings), 0)
         pb2.finished()
         pb1.finished()
 
@@ -160,28 +161,19 @@
         self.assertFalse(getattr(stack, 'note', False))
         pb1 = stack.get_nested()
         pb2 = stack.get_nested()
-        self.assertRaises(errors.MissingProgressBarFinish, pb1.finished)
+        warnings, _ = self.callCatchWarnings(pb1.finished)
+        self.assertEqual(len(warnings), 1)
         pb2.finished()
         pb1.finished()
         # the text ui factory never actually removes the stack once its setup.
         # we need to be able to nest again correctly from here.
         pb1 = stack.get_nested()
         pb2 = stack.get_nested()
-        self.assertRaises(errors.MissingProgressBarFinish, pb1.finished)
+        warnings, _ = self.callCatchWarnings(pb1.finished)
+        self.assertEqual(len(warnings), 1)
         pb2.finished()
         pb1.finished()
 
-    def test_text_factory_setting_progress_bar(self):
-        # we should be able to choose the progress bar type used.
-        factory = TextUIFactory(bar_type=DotsProgressBar)
-        bar = factory.nested_progress_bar()
-        bar.finished()
-        self.assertIsInstance(bar, DotsProgressBar)
-
-    def test_cli_stdin_is_default_stdin(self):
-        factory = CLIUIFactory()
-        self.assertEqual(sys.stdin, factory.stdin)
-
     def assert_get_bool_acceptance_of_user_input(self, factory):
         factory.stdin = StringIO("y\nyes with garbage\n"
                                  "yes\nn\nnot an answer\n"
@@ -212,28 +204,27 @@
         self.assertEqual('', factory.stdin.readline())
 
     def test_text_ui_getbool(self):
-        factory = TextUIFactory()
+        factory = TextUIFactory(None, None, None)
         self.assert_get_bool_acceptance_of_user_input(factory)
 
     def test_text_factory_prompts_and_clears(self):
         # a get_boolean call should clear the pb before prompting
-        factory = TextUIFactory(bar_type=DotsProgressBar)
-        factory.stdout = _TTYStringIO()
-        factory.stdin = StringIO("yada\ny\n")
-        pb = self.apply_redirected(factory.stdin, factory.stdout,
-                                   factory.stdout, factory.nested_progress_bar)
-        pb.start_time = None
-        self.apply_redirected(factory.stdin, factory.stdout,
-                              factory.stdout, pb.update, "foo", 0, 1)
+        out = _TTYStringIO()
+        factory = TextUIFactory(stdin=StringIO("yada\ny\n"), stdout=out, stderr=out)
+        pb = factory.nested_progress_bar()
+        pb.show_bar = False
+        pb.show_spinner = False
+        pb.show_count = False
+        pb.update("foo", 0, 1)
         self.assertEqual(True,
                          self.apply_redirected(None, factory.stdout,
                                                factory.stdout,
                                                factory.get_boolean,
                                                "what do you want"))
-        output = factory.stdout.getvalue()
-        self.assertEqual("foo: .\n"
-                         "what do you want? [y/n]: what do you want? [y/n]: ",
-                         factory.stdout.getvalue())
-        # stdin should be empty
+        output = out.getvalue()
+        self.assertContainsRe(factory.stdout.getvalue(),
+            "foo *\r\r  *\r*")
+        self.assertContainsRe(factory.stdout.getvalue(),
+            r"what do you want\? \[y/n\]: what do you want\? \[y/n\]: ")
+        # stdin should have been totally consumed
         self.assertEqual('', factory.stdin.readline())
-

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2008-12-12 13:09:26 +0000
+++ bzrlib/transport/__init__.py	2008-12-24 07:37:37 +0000
@@ -41,6 +41,7 @@
     errors,
     osutils,
     symbol_versioning,
+    ui,
     urlutils,
     )
 """)
@@ -382,6 +383,18 @@
         except TypeError: # We can't tell how many, because relpaths is a generator
             return None
 
+    def _report_activity(self, bytes, direction):
+        """Notify that this transport has activity.
+
+        Implementations should call this from all methods that actually do IO.
+        Be careful that it's not called twice, if one method is implemented on
+        top of another.
+
+        :param bytes: Number of bytes read or written.
+        :param direction: 'read' or 'write' or None.
+        """
+        ui.ui_factory.report_transport_activity(self, bytes, direction)
+
     def _update_pb(self, pb, msg, count, total):
         """Update the progress bar based on the current count
         and total available, total may be None if it was
@@ -568,7 +581,11 @@
 
         :param relpath: The relative path to the file
         """
-        return self.get(relpath).read()
+        f = self.get(relpath)
+        try:
+            return f.read()
+        finally:
+            f.close()
 
     @deprecated_method(one_four)
     def get_smart_client(self):

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2008-11-07 05:39:09 +0000
+++ bzrlib/transport/sftp.py	2009-01-15 05:12:49 +0000
@@ -1,5 +1,4 @@
-# Copyright (C) 2005 Robey Pointer <robey@lag.net>
-# Copyright (C) 2005, 2006, 2007 Canonical Ltd
+# Copyright (C) 2005, 2006, 2007, 2008, 2009 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -140,15 +139,18 @@
     # See _get_requests for an explanation.
     _max_request_size = 32768
 
-    def __init__(self, original_offsets, relpath):
+    def __init__(self, original_offsets, relpath, _report_activity):
         """Create a new readv helper.
 
         :param original_offsets: The original requests given by the caller of
             readv()
         :param relpath: The name of the file (if known)
+        :param _report_activity: A Transport._report_activity bound method,
+            to be called as data arrives.
         """
         self.original_offsets = list(original_offsets)
         self.relpath = relpath
+        self._report_activity = _report_activity
 
     def _get_requests(self):
         """Break up the offsets into individual requests over sftp.
@@ -219,6 +221,7 @@
             if len(data) != length:
                 raise errors.ShortReadvError(self.relpath,
                     start, length, len(data))
+            self._report_activity(length, 'read')
             if last_end is None:
                 # This is the first request, just buffer it
                 buffered_data = [data]
@@ -405,8 +408,7 @@
             return False
 
     def get(self, relpath):
-        """
-        Get the file at the given relative path.
+        """Get the file at the given relative path.
 
         :param relpath: The relative path to the file
         """
@@ -420,6 +422,16 @@
             self._translate_io_exception(e, path, ': error retrieving',
                 failure_exc=errors.ReadError)
 
+    def get_bytes(self, relpath):
+        # reimplement this here so that we can report how many bytes came back
+        f = self.get(relpath)
+        try:
+            bytes = f.read()
+            self._report_activity(len(bytes), 'read')
+            return bytes
+        finally:
+            f.close()
+
     def _readv(self, relpath, offsets):
         """See Transport.readv()"""
         # We overload the default readv() because we want to use a file
@@ -454,7 +466,7 @@
         does not support ranges > 64K, so it caps the request size, and
         just reads until it gets all the stuff it wants
         """
-        helper = _SFTPReadvHelper(offsets, relpath)
+        helper = _SFTPReadvHelper(offsets, relpath, self._report_activity)
         return helper.request_and_yield_offsets(fp)
 
     def put_file(self, relpath, f, mode=None):

=== modified file 'bzrlib/ui/__init__.py'
--- bzrlib/ui/__init__.py	2008-03-11 22:15:16 +0000
+++ bzrlib/ui/__init__.py	2009-01-13 05:07:27 +0000
@@ -26,7 +26,9 @@
 displays no output.
 """
 
+import os
 import sys
+import warnings
 
 from bzrlib.lazy_import import lazy_import
 lazy_import(globals(), """
@@ -49,8 +51,7 @@
     """
 
     def __init__(self):
-        super(UIFactory, self).__init__()
-        self._progress_bar_stack = None
+        self._task_stack = []
 
     def get_password(self, prompt='', **kwargs):
         """Prompt the user for a password.
@@ -73,7 +74,18 @@
         When the bar has been finished with, it should be released by calling
         bar.finished().
         """
-        raise NotImplementedError(self.nested_progress_bar)
+        if self._task_stack:
+            t = progress.ProgressTask(self._task_stack[-1], self)
+        else:
+            t = progress.ProgressTask(None, self)
+        self._task_stack.append(t)
+        return t
+
+    def progress_finished(self, task):
+        if task != self._task_stack[-1]:
+            warnings.warn("%r is not currently active" % (task,))
+        else:
+            del self._task_stack[-1]
 
     def clear_term(self):
         """Prepare the terminal for output.
@@ -104,13 +116,26 @@
             current_format_name,
             basedir)
 
+    def report_transport_activity(self, transport, byte_count, direction):
+        """Called by transports as they do IO.
+        
+        This may update a progress bar, spinner, or similar display.
+        By default it does nothing.
+        """
+        pass
+
+
 
 class CLIUIFactory(UIFactory):
-    """Common behaviour for command line UI factories."""
+    """Common behaviour for command line UI factories.
+    
+    This is suitable for dumb terminals that can't repaint existing text."""
 
-    def __init__(self):
-        super(CLIUIFactory, self).__init__()
-        self.stdin = sys.stdin
+    def __init__(self, stdin=None, stdout=None, stderr=None):
+        UIFactory.__init__(self)
+        self.stdin = stdin or sys.stdin
+        self.stdout = stdout or sys.stdout
+        self.stderr = stderr or sys.stderr
 
     def get_boolean(self, prompt):
         self.clear_term()
@@ -148,6 +173,15 @@
     def prompt(self, prompt):
         """Emit prompt on the CLI."""
 
+    def clear_term(self):
+        pass
+
+    def show_progress(self, task):
+        pass
+
+    def progress_finished(self, task):
+        pass
+
 
 class SilentUIFactory(CLIUIFactory):
     """A UI Factory which never prints anything.
@@ -155,19 +189,14 @@
     This is the default UI, if another one is never registered.
     """
 
+    def __init__(self):
+        CLIUIFactory.__init__(self)
+
     def get_password(self, prompt='', **kwargs):
         return None
 
-    def nested_progress_bar(self):
-        if self._progress_bar_stack is None:
-            self._progress_bar_stack = progress.ProgressBarStack(
-                klass=progress.DummyProgress)
-        return self._progress_bar_stack.get_nested()
-
-    def clear_term(self):
-        pass
-
-    def recommend_upgrade(self, *args):
+
+    def note(self, msg):
         pass
 
 
@@ -180,3 +209,23 @@
 ui_factory = SilentUIFactory()
 """IMPORTANT: never import this symbol directly. ONLY ever access it as 
 ui.ui_factory."""
+
+
+def make_ui_for_terminal(stdin, stdout, stderr):
+    """Construct and return a suitable UIFactory for a text mode program.
+
+    If stdout is a smart terminal, this gets a smart UIFactory with 
+    progress indicators, etc.  If it's a dumb terminal, just plain text output.
+    """
+    isatty = getattr(stdin, 'isatty', None)
+    if isatty is None:
+        cls = CLIUIFactory
+    elif not isatty():
+        cls = CLIUIFactory
+    elif os.environ.get('TERM') in (None, 'dumb', ''):
+        # e.g. emacs compile window
+        cls = CLIUIFactory
+    else:
+        from bzrlib.ui.text import TextUIFactory
+        cls = TextUIFactory
+    return cls(stdin=stdin, stdout=stdout, stderr=stderr)

=== modified file 'bzrlib/ui/text.py'
--- bzrlib/ui/text.py	2007-08-15 04:33:34 +0000
+++ bzrlib/ui/text.py	2008-12-16 07:56:29 +0000
@@ -1,4 +1,4 @@
-# Copyright (C) 2005 Canonical Ltd
+# Copyright (C) 2005, 2008 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,6 +20,7 @@
 """
 
 import sys
+import time
 
 from bzrlib.lazy_import import lazy_import
 lazy_import(globals(), """
@@ -28,7 +29,9 @@
 from bzrlib import (
     progress,
     osutils,
+    symbol_versioning,
     )
+
 """)
 
 from bzrlib.ui import CLIUIFactory
@@ -39,47 +42,189 @@
 
     def __init__(self,
                  bar_type=None,
+                 stdin=None,
                  stdout=None,
                  stderr=None):
         """Create a TextUIFactory.
 
         :param bar_type: The type of progress bar to create. It defaults to 
                          letting the bzrlib.progress.ProgressBar factory auto
-                         select.
+                         select.   Deprecated.
         """
-        super(TextUIFactory, self).__init__()
-        self._bar_type = bar_type
-        if stdout is None:
-            self.stdout = sys.stdout
-        else:
-            self.stdout = stdout
-        if stderr is None:
-            self.stderr = sys.stderr
-        else:
-            self.stderr = stderr
+        super(TextUIFactory, self).__init__(stdin=stdin,
+                stdout=stdout, stderr=stderr)
+        if bar_type:
+            symbol_versioning.warn(symbol_versioning.deprecated_in((1, 11, 0))
+                % "bar_type parameter")
+        # paints progress, network activity, etc
+        self._progress_view = TextProgressView(self.stderr)
 
     def prompt(self, prompt):
         """Emit prompt on the CLI."""
         self.stdout.write(prompt)
         
-    def nested_progress_bar(self):
-        """Return a nested progress bar.
-        
-        The actual bar type returned depends on the progress module which
-        may return a tty or dots bar depending on the terminal.
-        """
-        if self._progress_bar_stack is None:
-            self._progress_bar_stack = progress.ProgressBarStack(
-                klass=self._bar_type)
-        return self._progress_bar_stack.get_nested()
-
     def clear_term(self):
         """Prepare the terminal for output.
 
         This will, clear any progress bars, and leave the cursor at the
         leftmost position."""
-        if self._progress_bar_stack is None:
+        # XXX: If this is preparing to write to stdout, but that's for example
+        # directed into a file rather than to the terminal, and the progress
+        # bar _is_ going to the terminal, we shouldn't need
+        # to clear it.  We might need to separately check for the case of 
+        self._progress_view.clear()
+
+    def note(self, msg):
+        """Write an already-formatted message, clearing the progress bar if necessary."""
+        self.clear_term()
+        self.stdout.write(msg + '\n')
+
+    def report_transport_activity(self, transport, byte_count, direction):
+        """Called by transports as they do IO.
+        
+        This may update a progress bar, spinner, or similar display.
+        By default it does nothing.
+        """
+        self._progress_view.show_transport_activity(byte_count)
+
+    def show_progress(self, task):
+        """A task has been updated and wants to be displayed.
+        """
+        self._progress_view.show_progress(task)
+
+    def progress_finished(self, task):
+        CLIUIFactory.progress_finished(self, task)
+        if not self._task_stack:
+            # finished top-level task
+            self._progress_view.clear()
+
+
+class TextProgressView(object):
+    """Display of progress bar and other information on a tty.
+    
+    This shows one line of text, including possibly a network indicator, spinner, 
+    progress bar, message, etc.
+
+    One instance of this is created and held by the UI, and fed updates when a
+    task wants to be painted.
+
+    Transports feed data to this through the ui_factory object.
+    """
+
+    def __init__(self, term_file):
+        self._term_file = term_file
+        # true when there's output on the screen we may need to clear
+        self._have_output = False
+        # XXX: We could listen for SIGWINCH and update the terminal width...
+        self._width = osutils.terminal_width()
+        self._last_transport_msg = ''
+        self._spin_pos = 0
+        # time we last repainted the screen
+        self._last_repaint = 0
+        # time we last got information about transport activity
+        self._transport_update_time = 0
+        self._task_fraction = None
+        self._last_task = None
+        self._total_byte_count = 0
+        self._bytes_since_update = 0
+
+    def _show_line(self, s):
+        n = self._width - 1
+        self._term_file.write('\r%-*.*s\r' % (n, n, s))
+
+    def clear(self):
+        if self._have_output:
+            self._show_line('')
+        self._have_output = False
+
+    def _render_bar(self):
+        # return a string for the progress bar itself
+        if (self._last_task is not None) and self._last_task.show_bar:
+            spin_str =  r'/-\|'[self._spin_pos % 4]
+            self._spin_pos += 1
+            f = self._task_fraction or 0
+            cols = 20
+            # number of markers highlighted in bar
+            markers = int(round(float(cols) * f)) - 1
+            bar_str = '[' + ('#' * markers + spin_str).ljust(cols) + '] '
+            return bar_str
+        elif (self._last_task is None) or self._last_task.show_spinner:
+            spin_str =  r'/-\|'[self._spin_pos % 4]
+            self._spin_pos += 1
+            return spin_str + ' '
+        else:
+            return ''
+
+    def _format_task(self, task):
+        if not task.show_count:
+            s = ''
+        elif task.total_cnt is not None:
+            s = ' %d/%d' % (task.current_cnt, task.total_cnt)
+        elif task.current_cnt is not None:
+            s = ' %d' % (task.current_cnt)
+        else:
+            s = ''
+        self._task_fraction = task._overall_completion_fraction()
+        # compose all the parent messages
+        t = task
+        m = task.msg
+        while t._parent_task:
+            t = t._parent_task
+            if t.msg:
+                m = t.msg + ':' + m
+        return m + s
+
+    def _repaint(self):
+        bar_string = self._render_bar()
+        if self._last_task:
+            task_msg = self._format_task(self._last_task)
+        else:
+            task_msg = ''
+        trans = self._last_transport_msg
+        if trans and task_msg:
+            trans += ' | '
+        s = (bar_string
+             + trans
+             + task_msg
+             )
+        self._show_line(s)
+        self._have_output = True
+
+    def show_progress(self, task):
+        self._last_task = task
+        now = time.time()
+        if now < self._last_repaint + 0.1:
             return
-        overall_pb = self._progress_bar_stack.bottom()
-        if overall_pb is not None:
-            overall_pb.clear()
+        if now > self._transport_update_time + 5:
+            # no recent activity; expire it
+            self._last_transport_msg = ''
+        self._last_repaint = now
+        self._repaint()
+
+    def show_transport_activity(self, byte_count):
+        """Called by transports as they do IO.
+        
+        This may update a progress bar, spinner, or similar display.
+        By default it does nothing.
+        """
+        # XXX: Probably there should be a transport activity model, and that
+        # too should be seen by the progress view, rather than being poked in
+        # here.
+        self._total_byte_count += byte_count
+        self._bytes_since_update += byte_count
+        now = time.time()
+        if self._transport_update_time is None:
+            self._transport_update_time = now
+        elif now >= (self._transport_update_time + 0.2):
+            # guard against clock stepping backwards, and don't update too
+            # often
+            rate = self._bytes_since_update / (now - self._transport_update_time)
+            msg = ("%6dkB @ %4dkB/s" %
+                (self._total_byte_count>>10, int(rate)>>10,))
+            self._transport_update_time = now
+            self._last_repaint = now
+            self._bytes_since_update = 0
+            self._last_transport_msg = msg
+            self._repaint()
+
+

